/**
 * \file shuffle.c
 * \brief Shuffle Effect Demo - splits screen into tiles and shuffles them
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000-2002
 * by Thomas Nussbaumer & Lionel Debroux, with two modifications:
 * - this version uses a LFSR (pseudo-RNG) for scrambling the screen,
 *   instead of a larger and slower RNG;
 * - this version does not scramble adjacent blocks
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>
#include <libndls.h>
#include <stdint.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


#define BITS_PER_PIXEL  (4)
#define BITS_PER_BYTE   (8)

#define TILE_WIDTH (8)
#define TILE_HEIGHT (8)


//! Generator for 11-bit numbers.
#define LFSR_XOR_MASK (0x0500)


void Tile8x8Get_R(uint32_t col, uint32_t y, const void *src, uint32_t *dest);
void Tile8x8_RPLC_R(uint32_t col, uint32_t y, const uint32_t *sprite, void *plane);


int main(void) {
    void * screensave;
    uint32_t sprt1[8];
    uint32_t sprt2[8];
    uint32_t seq = 1234567891;

    show_msgbox(
        "shuffle - Shuffle effect demo",
        "------------------------\n"
        "   Shuffle effect demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "BKSPC - Reset the screen\n"
        "ESC   - Exit"
    );

    screensave = malloc(SCREEN_BYTES_SIZE);
    if (screensave != NULL) {
        // Save the screen.
        memcpy(screensave, SCREEN_BASE_ADDRESS, SCREEN_BYTES_SIZE);
        
        while (!isKeyPressed(KEY_NSPIRE_ESC)) {
            uint32_t x1, y1, x2, y2;

            // Simple LFSR.
            x1 = (seq / (SCREEN_WIDTH  / TILE_WIDTH));
            y1 = (seq % (SCREEN_HEIGHT / TILE_HEIGHT)) * TILE_HEIGHT;
            if (seq & 1) seq = (seq >> 1) ^ LFSR_XOR_MASK;
            else         seq = seq >> 1;

            x2 = (seq / (SCREEN_WIDTH  / TILE_WIDTH));
            y2 = (seq % (SCREEN_HEIGHT / TILE_HEIGHT)) * TILE_HEIGHT;
            if (seq & 1) seq = (seq >> 1) ^ LFSR_XOR_MASK;
            else         seq = seq >> 1;

            // It's cooler not to write outside of the bounds of the screen...
            if (   (x1 < SCREEN_WIDTH / TILE_WIDTH)
                && (x2 < SCREEN_WIDTH / TILE_WIDTH)
               ) {
                Tile8x8Get_R(x1, y1, SCREEN_BASE_ADDRESS, sprt1);
                Tile8x8Get_R(x2, y2, SCREEN_BASE_ADDRESS, sprt2);

                Tile8x8_RPLC_R(x1, y1, sprt2, SCREEN_BASE_ADDRESS);
                Tile8x8_RPLC_R(x2, y2, sprt1, SCREEN_BASE_ADDRESS);
            }


            WAIT("0x3FFF", "slowdown");

            if (isKeyPressed(KEY_NSPIRE_CLEAR)) {
                // Reinitialize the screen.
                memcpy(SCREEN_BASE_ADDRESS, screensave, SCREEN_BYTES_SIZE);
            }
        }
    }

    return 0;
}
